#!/usr/bin/env python

""" Analysis configuration base module.

For usage information, see ``jet_hadron.base.analysis_config``.

.. codeauthor:: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

import copy
import dataclasses
import enum
import itertools
import logging
import string
import ruamel.yaml
from typing import Any, Dict, List, Tuple, Type

logger = logging.getLogger(__name__)
# Make it a bit easier to specify the CommentedMap type.
DictLike = Type[ruamel.yaml.comments.CommentedMap]

def load_configuration(yaml: "ruamel.yaml.YAML", filename: str) -> DictLike:
    """ Load an analysis configuration from a file.

    Args:
        yaml: YAML object to use in loading the configuration.
        filename: Filename of the YAML configuration file.
    Returns:
        dict-like object containing the loaded configuration
    """
    with open(filename, "r") as f:
        config = yaml.load(f)

    return config

def override_options(config: DictLike, selected_options: tuple, set_of_possible_options: tuple, config_containing_override: DictLike = None) -> DictLike:
    """ Determine override options for a particular configuration.

    The options are determined by searching following the order specified in selected_options.

    For the example config,

    .. code-block:: yaml

        config:
            value: 3
            override:
                2.76:
                    track:
                        value: 5

    value will be assigned the value 5 if we are at 2.76 TeV with a track bias, regardless of the event
    activity or leading hadron bias. The order of this configuration is specified by the order of the
    selected_options passed. The above example configuration is from the jet-hadron analysis.

    Since anchors aren't kept for scalar values, if you want to override an anchored value, you need to
    specify it as a single value in a list (or dict, but list is easier). After the anchor values propagate,
    single element lists can be converted into scalar values using ``simplify_data_representations()``.

    Args:
        config: The dict-like configuration from ruamel.yaml which should be overridden.
        selected_options: The selected analysis options. They will be checked in the order with which
            they are passed, so make certain that it matches the order in the configuration file!
        set_of_possible_options (tuple of enums): Possible options for the override value categories.
        config_containing_override: The dict-like config containing the override options in a map called
            "override". If it is not specified, it will look for it in the main config.
    Returns:
        dict-like object: The updated configuration
    """
    if config_containing_override is None:
        config_containing_override = config
    override_opts = config_containing_override.pop("override")
    override_dict = determine_override_options(selected_options, override_opts, set_of_possible_options)
    logger.debug(f"override_dict: {override_dict}")

    # Set the configuration values to those specified in the override options
    # Cannot just use update() on config because we need to maintain the anchors.
    for k, v in override_dict.items():
        # Check if key is there and if it is not None! (The second part is important)
        if k in config:
            try:
                # If it has an anchor, we know that we want to preserve the type. So we check for the anchor
                # by trying to access it (Note that we don't actually care what the value is - just that it
                # exists). If it fails with an AttributeError, then we know we can just assign the value. If it
                # has an anchor, then we want to preserve the anchor information.
                config[k].anchor
                logger.debug(f"type: {type(config[k])}, k: {k}")
                if isinstance(config[k], list):
                    # Clear out the existing list entries
                    del config[k][:]
                    if isinstance(override_dict[k], (str, int, float, bool)):
                        # We have to treat str carefully because it is an iterable, but it will be expanded as
                        # individual characters if it's treated the same as a list, which is not the desired
                        # behavior! If we wrap it in [], then it will be treated as the only entry in the list
                        # NOTE: We also treat the basic types this way because they will be passed this way if
                        #       overriding indirectly with anchors (since the basic scalar types don't yet support
                        #       reassignment while maintaining their anchors).
                        config[k].append(override_dict[k])
                    else:
                        # Here we just assign all entries of the list to all entries of override_dict[k]
                        config[k].extend(override_dict[k])
                elif isinstance(config[k], dict):
                    # Clear out the existing entries because we are trying to replace everything
                    # Then we can simply update the dict with our new values
                    config[k].clear()
                    config[k].update(override_dict[k])
                elif isinstance(config[k], (int, float, bool)):
                    # This isn't really very good (since we lose information), but there's nothing that can be done
                    # about it at the moment (Dec 2018)
                    logger.debug("Overwriting YAML anchor object. It is currently unclear how to reassign this value.")
                    config[k] = v
                else:
                    # Raise a value error on all of the cases that we aren't already aware of.
                    raise ValueError(f"Object {k} (type {type(config[k])}) somehow has an anchor, but is something other than a list or dict. Attempting to assign directly to it.")
            except AttributeError:
                # If no anchor, just overwrite the value at this key
                config[k] = v
        else:
            raise KeyError(k, f"Trying to override key \"{k}\" that it is not in the config.")

    return config

def simplify_data_representations(config: DictLike) -> DictLike:
    """ Convert one entry lists to the scalar value

    This step is necessary because anchors are not kept for scalar values - just for lists and dictionaries.
    Now that we are done with all of our anchor references, we can convert these single entry lists to
    just the scalar entry, which is more usable.

    Some notes on anchors in ruamel.yaml are here: https://stackoverflow.com/a/48559644

    Args:
        config: The dict-like configuration from ruamel.yaml which should be simplified.
    Returns:
        The updated configuration.
    """
    for k, v in config.items():
        if v and isinstance(v, list) and len(v) == 1:
            logger.debug("v: {}".format(v))
            config[k] = v[0]

    return config

def determine_override_options(selected_options: tuple, override_opts: DictLike, set_of_possible_options: tuple = ()):
    """ Recursively extract the dict described in override_options().

    In particular, this searches for selected options in the override_opts dict. It stores only
    the override options that are selected.

    Args:
        selected_options: The options selected for this analysis, in the order defined used
            with ``override_options()`` and in the configuration file.
        override_opts: dict-like object returned by ruamel.yaml which contains the options that
            should be used to override the configuration options.
        set_of_possible_options (tuple of enums): Possible options for the override value categories.
    """
    override_dict: Dict[str, Any] = {}
    for option in override_opts:
        # We need to cast the option to a string to effectively compare to the selected option,
        # since only some of the options will already be strings
        if str(option) in list(map(lambda opt: str(opt), selected_options)):
            override_dict.update(determine_override_options(selected_options, override_opts[option], set_of_possible_options))
        else:
            logger.debug(f"override_opts: {override_opts}")
            # Look for whether the key is one of the possible but unselected options.
            # If so, we haven't selected it for this analysis, and therefore they should be ignored.
            # NOTE: We compare both the names and value because sometimes the name is not sufficient,
            #       such as in the case of the energy (because a number is not allowed to be a field name.)
            found_as_possible_option = False
            for possible_options in set_of_possible_options:
                # Same type of comparison as above, but for all possible options instead of the selected
                # options.
                if str(option) in list(map(lambda opt: str(opt), possible_options)):
                    found_as_possible_option = True
                # Below is more or less equivalent to the above (although .str() hides the details or
                # whether we should compare to the name or the value in the enum and only compares against
                # the designated value).
                #for possible_opt in possible_options:
                    #if possible_opt.name == option or possible_opt.value == option:
                    #    found_as_possible_option = True

            if not found_as_possible_option:
                # Store the override value, since it doesn't correspond with a selected option or a possible
                # option and therefore must be an option that we want to override.
                logger.debug(f"Storing override option \"{option}\", with value \"{override_opts[option]}\"")
                override_dict[option] = override_opts[option]
            else:
                logger.debug(f"Found option \"{option}\" as possible option, so skipping!")

    return override_dict

def determine_selection_of_iterable_values_from_config(config: DictLike, possible_iterables: Dict[str, Type[enum.Enum]]) -> Dict[str, List[Any]]:
    """ Determine iterable values to use to create objects for a given configuration.

    All values of an iterable can be included be setting the value to ``True`` (Not as a single value list,
    but as the only value.). Alternatively, an iterator can be disabled by setting the value to ``False``.

    Args:
        config: The dict-like configuration from ruamel.yaml which should be overridden.
        possible_iterables: Key value pairs of names of enumerations and their values.
    Returns:
        dict: Iterables values that were requested in the config.
    """
    iterables = {}
    requested_iterables = config["iterables"]
    for k, v in requested_iterables.items():
        if k not in possible_iterables:
            raise KeyError(k, f"Cannot find requested iterable in possible_iterables: {possible_iterables}")
        logger.debug(f"k: {k}, v: {v}")
        additional_iterable: List[Any] = []
        enum_values = possible_iterables[k]
        # Check for a string. This is wrong, and the user should be notified.
        if isinstance(v, str):
            raise TypeError(type(v), f"Passed string {v} when must be either bool or list")
        # Allow the possibility to skip
        if v is False:
            continue
        # Allow the possibility to including all possible values in the enum.
        elif v is True:
            additional_iterable = list(enum_values)
        else:
            if enum_values is None:
                # The enumeration values are none, which means that we want to take
                # all of the values defined in the config.
                additional_iterable = list(v)
            else:
                # Otherwise, only take the requested values.
                for el in v:
                    additional_iterable.append(enum_values[el])
        # Store for later
        iterables[k] = additional_iterable

    return iterables

def create_objects_from_iterables(obj, args: dict, iterables: dict, formatting_options: dict, key_index_name: str = "KeyIndex") -> Tuple[Any, Dict[str, Any], dict]:
    """ Create objects for each set of values based on the given arguments.

    The iterable values are available under a key index ``dataclass`` which is used to index the returned
    dictionary. The names of the fields are determined by the keys of iterables dictionary. The values are
    the newly created object. Note that the iterable values must be convertible to a str() so they can be
    included in the formatting dictionary.

    Each set of values is also included in the object args.

    As a basic example,

    .. code-block:: python

        >>> create_objects_from_iterables(
        ...     obj = obj,
        ...     args = {},
        ...     iterables = {"a" : ["a1","a2"], "b" : ["b1", "b2"]},
        ...     formatting_options = {}
        ... )
        (
            KeyIndex,
            {"a": ["a1", "a2"], "b": ["b1", "b2"]}
            {
                KeyIndex(a = "a1", b = "b1"): obj(a = "a1", b = "b1"),
                KeyIndex(a = "a1", b = "b2"): obj(a = "a1", b = "b2"),
                KeyIndex(a = "a2", b = "b1"): obj(a = "a2", b = "b1"),
                KeyIndex(a = "a2", b = "b2"): obj(a = "a2", b = "b2"),
            }
        )

    Args:
        obj (object): The object to be constructed.
        args: Arguments to be passed to the object to create it.
        iterables: Iterables to be used to create the objects, with entries of the form
            ``"name_of_iterable": iterable``.
        formatting_options (dict): Values to be used in formatting strings in the arguments.
        key_obj_name (str): Name of the iterable key object.
    Returns:
        (object, list, dict, dict): Roughly, (KeyIndex, iterables, objects). Specifically, the
            key_index is a new dataclass which defines the parameters used to create the object, iterables
            are the iterables used to create the objects, which names as keys and the iterables as values.
            The objects dictionary keys are KeyIndex objects which describe the iterable arguments passed to the
            object, while the values are the newly constructed arguments. See the example above.
    """
    # Setup
    objects = {}
    names = list(iterables)
    logger.debug("iterables: {iterables}".format(iterables = iterables))
    # Create the key index object, where the name of each field is the name of each iterable.
    KeyIndex = dataclasses.make_dataclass(
        key_index_name,
        [(name, type(iterable)) for name, iterable in iterables.items()],
        frozen = True
    )
    # ``itertools.product`` produces all possible permutations of the iterables values.
    # NOTE: Product preserves the order of the iterables values, which is important for properly
    #       assigning the values to the ``KeyIndex``.
    for values in itertools.product(*iterables.values()):
        logger.debug(f"Values: {values}")
        # Skip if we don't have a sufficient set of values to create an object.
        if not values:
            continue

        # Add in the values into the arguments and formatting options.
        # NOTE: We don't need a deep copy for the iterable values in the args and formatting options
        #       because the values will be overwritten for each object.
        for name, val in zip(names, values):
            # We want to keep the original value for the arguments.
            args[name] = val
            # Here, we convert the value, regardless of type, into a string that can be displayed.
            formatting_options[name] = str(val)

        # Apply formatting options
        # If we formatted in place, we would need to deepcopy the args to ensure that the iterable dependent
        # values in the formatted values are properly set for each iterable object individually.
        # However, by formatting into new variables, we can avoid a deepcopy, which greatly improves performance!
        # NOTE: We don't need a deep copy do this for iterable value names themselves because they will be overwritten
        #       for each object. They are set in the block above.
        object_args = copy.copy(args)
        logger.debug(f"object_args pre format: {object_args}")
        object_args = apply_formatting_dict(object_args, formatting_options)
        # Print our results for debugging purposes. However, we skip printing the full
        # config because it is quite long
        print_args = {k: v for k, v in object_args.items() if k != "config"}
        print_args["config"] = "..."
        logger.debug(f"Constructing obj \"{obj}\" with args: \"{print_args}\"")

        # Finally create the object.
        objects[KeyIndex(*values)] = obj(**object_args)

    # If nothing has been created at this point, then we are didn't iterating over anything and something
    # has gone wrong.
    if not objects:
        raise ValueError(iterables, "There appear to be no iterables to use in creating objects.")

    return (KeyIndex, iterables, objects)

class formatting_dict(dict):
    """ Dict to handle missing keys when formatting a string.

    It returns the missing key for later use in formatting. See: https://stackoverflow.com/a/17215533
    """
    def __missing__(self, key: str) -> str:
        return "{" + key + "}"

def apply_formatting_dict(obj: Any, formatting: Dict[str, Any]) -> Any:
    """ Recursively apply a formatting dict to all strings in a configuration.

    Note that it skips applying the formatting if the string appears to contain latex (specifically,
    if it contains an "$"), since the formatting fails on nested brackets.

    Args:
        obj: Some configuration object to recursively applying the formatting to.
        formatting (dict): String formatting options to apply to each configuration field.
    Returns:
        dict: Configuration with formatting applied to every field.
    """
    #logger.debug("Processing object of type {}".format(type(obj)))
    new_obj = obj

    if isinstance(obj, str):
        # Apply the formatting options to the string.
        # We explicitly allow for missing keys. They will be kept so they can be filled later.
        # see: https://stackoverflow.com/a/17215533
        # If a more sophisticated solution is needed,
        # see: https://ashwch.github.io/handling-missing-keys-in-str-format-map.html
        # Note that we can't use format_map because it is python 3.2+ only.
        # The solution below works in py 2/3
        if "$" not in obj:
            new_obj = string.Formatter().vformat(obj, (), formatting_dict(**formatting))
        #else:
        #    logger.debug("Skipping str {} since it appears to be a latex string, which may break the formatting.".format(obj))
    elif isinstance(obj, dict):
        new_obj = {}
        for k, v in obj.items():
            # Using indirect access to ensure that the original object is updated.
            new_obj[k] = apply_formatting_dict(v, formatting)
    elif isinstance(obj, list):
        new_obj = []
        for i, el in enumerate(obj):
            # Using indirect access to ensure that the original object is updated.
            new_obj.append(apply_formatting_dict(el, formatting))
    elif isinstance(obj, int) or isinstance(obj, float) or obj is None:
        # Skip over this, as there is nothing to be done - we just keep the value.
        pass
    elif isinstance(obj, enum.Enum):
        # Skip over this, as there is nothing to be done - we just keep the value.
        # This only occurs when a formatting value has already been transformed
        # into an enumeration.
        pass
    else:
        # This may or may not be expected, depending on the particular value.
        logger.info(f"NOTE: Unrecognized type {type(obj)} of obj {obj}")

    return new_obj

def iterate_with_selected_objects(analysis_objects: Dict[Any, Any], **selections: Dict[str, Any]) -> Any:
    """ Iterate over an analysis dictionary with selected attributes.

    Args:
        analysis_objects: Analysis objects dictionary.
        selections: Keyword arguments used to select attributes from the analysis dictionary.
    Yields:
        object: Matching analysis object.
    """
    for key_index, obj in analysis_objects.items():
        # If selections is empty, we return every object. If it's not empty, then we only want to return
        # objects which are selected in through the selections.
        selected_obj = not selections or all([getattr(key_index, selector) == selected_value for selector, selected_value in selections.items()])

        if selected_obj:
            yield key_index, obj




#!/usr/bin/env python

""" Broad collection of utility functions and constants.

.. codeauthor:: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

import logging
import numpy as np
import ruamel.yaml

from pachyderm import histogram

# Setup logger
logger = logging.getLogger(__name__)

# Small value - epsilon
# For use to offset from bin edges when finding bins for use with SetRange()
# NOTE: sys.float_info.epsilon is too small in some cases and thus should be avoided
epsilon = 1e-5

###################
# Utility functions
###################
def read_YAML(filename: str, file_access_mode: str = "r") -> ruamel.yaml.comments.CommentedMap:
    """ Read the YAML file at filename.

    Uses the round trip mode.

    Args:
        filename (str): Filename of the YAML file to be read.
        file_access_mode (str): Mode under which the file should be opened
    Returns:
        dict-like: Dict containing the parameters read from the YAML file.
    """
    parameters = None
    with open(filename, file_access_mode) as f:
        yaml = ruamel.yaml.YAML(typ = "rt")
        yaml.default_flow_style = False
        parameters = yaml.load(f)
    return parameters

def write_YAML(parameters: dict, filename: str, file_access_mode: str = "w") -> None:
    """ Write the given output dict to file using YAML.

    Uses the round trip mode.

    Args:
        parameters (dict): Output parameters to be written to the YAML file.
        filename (str): Filename of the YAML file to write.
        file_access_mode (str): Mode under which the file should be opened
    Returns:
        None.
    """
    with open(filename, file_access_mode) as f:
        yaml = ruamel.yaml.YAML(typ = "rt")
        yaml.default_flow_style = False
        yaml.dump(parameters, f)

def moving_average(arr: np.ndarray, n: int = 3) -> np.ndarray:
    """ Calculate the moving overage over an array.

    Algorithm from: https://stackoverflow.com/a/14314054

    Args:
        arr (np.ndarray): Array over which to calculate the moving average.
        n (int): Number of elements over which to calculate the moving average. Default: 3
    Returns:
        np.ndarray: Moving average calculated over n.
    """
    ret = np.cumsum(arr, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n

def get_array_for_fit(observables: dict, track_pt_bin: int, jet_pt_bin: int) -> histogram.Histogram1D:
    """ Get a Histogram1D associated with the selected jet and track pt bins.

    This is often used to retrieve data for fitting.

    Args:
        observables (dict): The observables from which the hist should be retrieved.
        track_pt_bin (int): Track pt bin of the desired hist.
        jet_ptbin (int): Jet pt bin of the desired hist.
    Returns:
        Histogram1D: Converted TH1 or uproot histogram.
    Raises:
        ValueError: If the requested observable couldn't be found.
    """
    for name, observable in observables.items():
        if observable.track_pt_bin == track_pt_bin and observable.jet_pt_bin == jet_pt_bin:
            return histogram.Histogram1D.from_existing_hist(observable.hist)

    raise ValueError("Cannot find fit with jet pt bin {jet_pt_bin} and track pt bin {track_pt_bin}")



#!/usr/bin/env python

""" Contains generic classes

.. codeauthor:: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

class EqualityMixin(object):
    """ Mixin generic comparison operations using `__dict__`.

    Can then be mixed into any other class using multiple inheritance.

    Inspired by: https://stackoverflow.com/a/390511.
    """
    def __eq__(self, other) -> bool:
        """ Check for equality of members. """
        # Check identity to avoid needing to perform the (potentially costly) dict comparison.
        if self is other:
            return True
        # Compare via the member values.
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return NotImplemented




#!/usr/bin/env python

""" Histogram related classes and functionality.

.. codeauthor:: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

from dataclasses import dataclass
import logging
import numpy as np
from typing import Any, Dict, Tuple

# Setup logger
logger = logging.getLogger(__name__)

def get_histograms_in_file(filename: str) -> Dict[str, Any]:
    """ Helper function which gets all histograms in a file.

    Args:
        filename: Filename of the ROOT file containing the list.
    Returns:
        Contains hists with keys as their names. Lists are recursively added, mirroring
            the structure under which the hists were stored.
    """
    return get_histograms_in_list(filename = filename)

def get_histograms_in_list(filename: str, list_name: str = None) -> Dict[str, Any]:
    """ Get histograms from the file and make them available in a dict.

    Lists are recursively explored, with all lists converted to dictionaries, such that the return
    dictionaries which only contains hists and dictionaries of hists (ie there are no ROOT ``TCollection``
    derived objects).

    Args:
        filename: Filename of the ROOT file containing the list.
        list_name: Name of the list to retrieve.
    Returns:
        Contains hists with keys as their names. Lists are recursively added, mirroring
            the structure under which the hists were stored.
    Raises:
        ValueError: If the list could not be found in the given file.
    """
    import ROOT

    hists: dict = {}
    fIn = ROOT.TFile(filename, "READ")
    if list_name is not None:
        hist_list = fIn.Get(list_name)
    else:
        hist_list = [obj.ReadObj() for obj in fIn.GetListOfKeys()]

    if not hist_list:
        fIn.ls()
        # Closing this file appears (but is not entirely confirmed) to be extremely important! Otherwise,
        # the memory will leak, leading to ROOT memory issues!
        fIn.Close()
        raise ValueError(f"Could not find list with name \"{list_name}\". Possible names are listed above.")

    # Retrieve objects in the hist list
    for obj in hist_list:
        _retrieve_object(hists, obj)

    # Cleanup
    fIn.Close()

    return hists

def _retrieve_object(output_dict: Dict[str, Any], obj: Any) -> None:
    """ Function to recursively retrieve histograms from a list in a ROOT file.

    ``SetDirectory(True)`` is applied to TH1 derived hists and python is explicitly given
    ownership of the retrieved objects.

    Args:
        output_dict (dict): Dict under which hists should be stored.
        obj (ROOT.TObject derived): Object(s) to be stored. If it is a collection,
            it will be recursed through.
    Returns:
        None: Changes in the dict are reflected in the output_dict which was passed.
    """
    import ROOT

    # Store TH1 or THn
    if isinstance(obj, ROOT.TH1) or isinstance(obj, ROOT.THnBase):
        # Ensure that it is not lost after the file is closed
        # Only works for TH1
        if isinstance(obj, ROOT.TH1):
            obj.SetDirectory(0)

        # Explicitly note that python owns the object
        # From more on memory management with ROOT and python, see:
        # https://root.cern.ch/root/html/guides/users-guide/PythonRuby.html#memory-handling
        ROOT.SetOwnership(obj, False)

        # Store the object
        output_dict[obj.GetName()] = obj

    # Recurse over lists
    if isinstance(obj, ROOT.TCollection):
        # Keeping it in order simply makes it easier to follow
        output_dict[obj.GetName()] = {}
        # Iterate over the objects in the collection and recursively store them
        for obj_temp in list(obj):
            _retrieve_object(output_dict[obj.GetName()], obj_temp)

@dataclass
class Histogram1D:
    """ Contains histogram data.

    Note:
        Underflow and overflow bins are excluded!

    Args:
        bin_edges (np.ndarray): The histogram bin edges.
        y (np.ndarray): The histogram bin values.
        errors_squared (np.ndarray): The bin sum weight squared errors.

    Attributes:
        x (np.ndarray): The bin centers.
        y (np.ndarray): The bin values.
        bin_edges (np.ndarray): The bin edges.
        errors (np.ndarray): The bin errors.
        errors_squared (np.ndarray): The bin sum weight squared errors.
    """
    bin_edges: np.ndarray
    y: np.ndarray
    errors_squared: np.ndarray

    @property
    def errors(self) -> np.ndarray:
        return np.sqrt(self.errors_squared)

    @property
    def x(self) -> np.ndarray:
        """ The histogram bin centers (``x``).

        This property caches the x value so we don't have to calculate it every time.
        """
        try:
            return self._x
        except AttributeError:
            bin_widths = (self.bin_edges[1:] - self.bin_edges[:-1]) / 2
            x = self.bin_edges[:-1] + bin_widths
            self._x: np.ndarray = x

        return self._x

    @staticmethod
    def _from_uproot(hist) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """ Convert a uproot histogram to a set of array for creating a Histogram.

        Note:
            Underflow and overflow bins are excluded!

        Args:
            hist (uproot.hist.TH1*): Input histogram.
        Returns:
            tuple: (x, y, errors) where x is the bin centers, y is the bin values, and
                errors are the sumw2 bin errors.
        """
        # This excluces underflow and overflow
        (y, bin_edges) = hist.numpy()

        # Also retrieve errors from sumw2.
        # If more sophistication is needed, we can modify this to follow the approach to
        # calculating bin errors from TH1::GetBinError()
        errors = hist.variances

        return (bin_edges, y, errors)

    @staticmethod
    def _from_th1(hist) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """ Convert a TH1 histogram to a Histogram.

        Note:
            Underflow and overflow bins are excluded!

        Args:
            hist (ROOT.TH1): Input histogram.
        Returns:
            tuple: (x, y, errors) where x is the bin centers, y is the bin values, and
                errors are the sumw2 bin errors.
        """
        # Enable sumw2 if it's not already calculated
        if hist.GetSumw2N() == 0:
            hist.Sumw2(True)

        # Don't include overflow
        bin_edges = get_bin_edges_from_axis(hist.GetXaxis())
        # NOTE: The y value and bin error are stored with the hist, not the axis.
        y = np.array([hist.GetBinContent(i) for i in range(1, hist.GetXaxis().GetNbins() + 1)])
        errors = np.array(hist.GetSumw2())
        # Exclude the under/overflow binsov
        errors = errors[1:-1]

        return (bin_edges, y, errors)

    @classmethod
    def from_existing_hist(cls, hist: Any):
        """ Convert an existing histogram.

        Note:
            Underflow and overflow bins are excluded! Bins are assumed to be fixed
            size.

        Args:
            hist (uproot.rootio.TH1* or ROOT.TH1): Histogram to be converted.
        Returns:
            Histogram: Dataclass with x, y, and errors
        """
        try:
            # Convert jet_hadron.base.analysis_objects.HistogramContainer -> TH1 or uproot hist.
            # It goes HistogramContainer.hist -> TH1 or uproot hist
            logger.debug("Converting HistogramContainer to standard hist")
            hist = hist.hist
        except AttributeError:
            # Just use the existing histogram
            pass
        # "values" is a proxy for if we have an uproot hist.
        logger.debug(f"{hist}, {type(hist)}")
        if hasattr(hist, "values"):
            (bin_edges, y, errors_squared) = cls._from_uproot(hist)
        else:
            # Handle traditional ROOT hists
            (bin_edges, y, errors_squared) = cls._from_th1(hist)

        return cls(bin_edges = bin_edges, y = y, errors_squared = errors_squared)

def get_array_from_hist2D(hist: Any, set_zero_to_NaN: bool = True, return_bin_edges: bool = False) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """ Extract x, y, and bin values from a 2D ROOT histogram.

    Converts the histogram into a numpy array, and suitably processes it for a surface plot
    by removing 0s (which can cause problems when taking logs), and returning a set of (x, y) mesh
    values utilziing either the bin edges or bin centers.

    Note:
        This is a different format than the 1D version!

    Args:
        hist (ROOT.TH2): Histogram to be converted.
        set_zero_to_NaN: If true, set 0 in the array to NaN. Useful with matplotlib so that it will
            ignore the values when plotting. See comments in this function for more details. Default: True.
        return_bin_edges: Return x and y using bin edges instead of bin centers.
    Returns:
        Contains (x values, y values, numpy array of hist data) where (x, y) are values on a
            grid (from np.meshgrid) using the selected bin values.
    """
    # Process the hist into a suitable state
    # NOTE: The shape specific can be somewhat confusing (ie. I would naviely expected to specify the x first.)
    # This says that the ``GetYaxis().GetNbins()`` number of rows and ``GetXaxis().GetNbins()`` number of columns.
    shape = (hist.GetYaxis().GetNbins(), hist.GetXaxis().GetNbins())
    # To keep consistency with the root_numpy 2D hist format, we transpose the final result
    # This format has x values as columns.
    hist_array = np.array([hist.GetBinContent(x) for x in range(1, hist.GetNcells()) if not hist.IsBinUnderflow(x) and not hist.IsBinOverflow(x)])
    # The hist_array was linear, so we need to shape it into our expected 2D values.
    hist_array = hist_array.reshape(shape)
    # Transpose the array to better match expectations
    # In particular, by transposing the array, it means that ``thist_array[1][0]`` gives the 2nd x
    # value (x_index = 1) and the 1st y value (y_index = 1). This is as we would expect. This is also
    # the same convention as used by root_numpy
    hist_array = hist_array.T
    # Set all 0s to nan to get similar behavior to ROOT. In ROOT, it will basically ignore 0s. This is
    # especially important for log plots. Matplotlib doesn't handle 0s as well, since it attempts to
    # plot them and then will throw exceptions when the log is taken.
    # By setting to nan, matplotlib basically ignores them similar to ROOT
    # NOTE: This requires a few special functions later which ignore nan when calculating min and max.
    if set_zero_to_NaN:
        hist_array[hist_array == 0] = np.nan

    if return_bin_edges:
        # Bin edges
        x_bin_edges = get_bin_edges_from_axis(hist.GetXaxis())
        y_bin_edges = get_bin_edges_from_axis(hist.GetYaxis())

        # NOTE: The addition of epsilon to the max is extremely important! Otherwise, the x and y
        #       ranges will be one bin short since ``arange`` is not inclusive. This could also be resolved
        #       by using ``linspace``, but I think this approach is perfectly fine.
        # NOTE: This epsilon is smaller than the one in ``utils`` because we are sometimes dealing
        #       with small times (~ns). The other value is larger because (I seem to recall) that
        #       smaller values didn't always place nice with ROOT, but it is fine here, since we're
        #       working with numpy.
        # NOTE: This should be identical to taking the min and max of the axis using
        #       ``TAxis.GetXmin()`` and ``TAxis.GetXmax()``, but I prefer this approach.
        epsilon = 1e-9
        x_range = np.arange(
            np.amin(x_bin_edges),
            np.amax(x_bin_edges) + epsilon,
            hist.GetXaxis().GetBinWidth(1)
        )
        y_range = np.arange(
            np.amin(y_bin_edges),
            np.amax(y_bin_edges) + epsilon,
            hist.GetYaxis().GetBinWidth(1)
        )
    else:
        # We want an array of bin centers
        x_range = np.array([hist.GetXaxis().GetBinCenter(i) for i in range(1, hist.GetXaxis().GetNbins() + 1)])
        y_range = np.array([hist.GetYaxis().GetBinCenter(i) for i in range(1, hist.GetYaxis().GetNbins() + 1)])

    X, Y = np.meshgrid(x_range, y_range)

    return (X, Y, hist_array)

def get_bin_edges_from_axis(axis) -> np.ndarray:
    """ Get bin edges from a ROOT hist axis.

    Note:
        Doesn't include over- or underflow bins!

    Args:
        axis (ROOT.TAxis): Axis from which the bin edges should be extracted.
    Returns:
        Array containing the bin edges.
    """
    # Don't include over- or underflow bins
    bins = range(1, axis.GetNbins() + 1)
    # Bin edges
    bin_edges = np.empty(len(bins) + 1)
    bin_edges[:-1] = [axis.GetBinLowEdge(i) for i in bins]
    bin_edges[-1] = axis.GetBinUpEdge(axis.GetNbins())

    return bin_edges




#/usr/bin/env python

""" Setup pachyderm.

Derived from the setup.py in aliBuild and Overwatch
and based on: https://python-packaging.readthedocs.io/en/latest/index.html

.. codeauthor:: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
import os

def get_version():
    versionModule = {}
    with open(os.path.join("pachyderm", "version.py")) as f:
        exec(f.read(), versionModule)
    return versionModule["__version__"]

# Get the long description from the README file
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="pachyderm",
    version=get_version(),

    description="Physics Analysis Core for Heavy-Ions",
    long_description=long_description,
    long_description_content_type="text/markdown",

    author="Raymond Ehlers",
    author_email="raymond.ehlers@cern.ch",

    url="https://github.com/raymondEhlers/pachyderm",
    license="BSD 3-Clause",

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],

    # What does your project relate to?
    keywords='HEP ALICE',

    packages=find_packages(exclude=(".git", "tests")),

    # Rename scripts to the desired executable names
    # See: https://stackoverflow.com/a/8506532
    entry_points = {
        "console_scripts": [
        ],
    },

    # This is usually the minimal set of the required packages.
    # Packages should be installed via pip -r requirements.txt !
    install_requires=[
        "dataclasses;python_version<'3.7'",
        "ruamel.yaml",
        "numpy",
        "uproot",
        # Depends on ROOT, but that can't be installed through pip.
        # The dependence is only implicit, so it can actually be installed with it,
        # but some functionality depends on it being available.
        #"ROOT"
    ],

    # Include additional files
    include_package_data=True,

    extras_require = {
        "tests": [
            "pytest",
            "pytest-cov",
            "pytest-mock",
            "codecov",
        ],
        "docs": [
            "sphinx",
            # Allow markdown files to be used
            "recommonmark",
            # Allow markdown tables to be used
            "sphinx_markdown_tables",
        ],
        "dev": [
            "pre-commit",
            "flake8",
            # Makes flake8 easier to parse
            "flake8-colors",
            # Type checking
            "mypy",
        ]
    }
)
